package contextG

import static org.junit.Assert.*

class FibMath {
	// preferred way: explicit layer defintion
	static Layer CountCalls = new Layer()
	
	def calls = -1
    def fib(n) {
        if(n < 2) return 1
        fib(n - 1) + fib(n - 2)
    }
    
    static {
        layer(CountCalls) {
            calls = 0
        	staticCalls = 0
            reset = {
            	staticCalls = calls = 0
            }
            fib = {n ->            
                proceed()
                ++calls
            }
        }
    }
    
    public static void main(String[] args) {
    	def m = new FibMath()
    	// regular result is 89 and calls is unmodified
    	assertEquals 89, m.fib(10) 
    	assertEquals(-1, m.calls)
    	
    	// activate count layer
    	with(FibMath.CountCalls) {
    		// to calculate fib(x) recursively, 2 * x - 1 calls are needed
    		assertEquals 177, m.fib(10)
    		
    		// reset calls in the context of the active layer (it is always -1 outside this layer)
            m.calls = 0
        	assertEquals 0, m.calls
        	
            // same result as above because calls was resetted
    		assertEquals 177, m.fib(10)
    		
    		// doubled result
    		assertEquals 354, m.fib(10)
        	assertEquals 354, m.calls
        }

    	// m.calls did never change outside the layer
    	assertEquals(-1, m.calls)
    	
    	// multiple activation...
    	with(FibMath.CountCalls) {
        	with(FibMath.CountCalls) {
        		// ...are counted as one activation
        		assertEquals 354, m.calls
        		
        		// deactivation has precedence
        		// the layer is not activated indepent of how many times it was activated before
        		without(FibMath.CountCalls) {
        	    	assertEquals(-1, m.calls)
        			
        			// but if it is activated again it is active
        			with(FibMath.CountCalls) {
                		assertEquals 354, m.calls
        			}

        	    	assertEquals(-1, m.calls)
        		}
        	}
        	// multiple activations ensure that the layer is active until the last activation ended
    		assertEquals 354, m.calls
    	}

    	// all activations ended
    	assertEquals(-1, m.calls)
    	
    	// methods can be added in layers
    	try {
    		m.reset()
    		fail()
    	} catch(MissingMethodException) {}
    	
    	with(FibMath.CountCalls) {
    		m.reset()
        	assertEquals 0, m.calls
    	}
    }
}
