package contextG

import static org.junit.Assert.*

class Employer {  
    String name
    String address
  
    String toString() {
        "Name: " + name
    }
    
    static {
        layer(Layer.Address) {
            toString = {
        		// proceed() calls the method which would have been called if this layer was not active
        		// this could be the original definition of the method or 
        		// a redefined method of a previously activated layer
                proceed() + "; Address: " + address
            }
        }
    }
}

class Person {    
    String name
    String address
    Employer employer
    
    String toString() {
        "Name: " + name
    }
    
    static {
        layer(Layer.Address) {
            toString = {
                proceed() + "; Address: " + address
            }
        }
        layer(Layer.Employment) {
            toString = {
                proceed() + "; [Employer] " + employer
            }
        }
    }
    
    public static void main(String[] args) {
        new Person().test()
    }
    
    def inspect(o) {
        println "${Layer.activeLayers.get()} -> ${o.toString()}"
    }
    
    def test() {
        Employer HPI = new Employer(name: "HPI", address: "14482 Potsdam")
        // without layer
        assertEquals "Name: HPI", HPI.toString()
        // now add address
        with (Layer.Address) {
        	assertEquals "Name: HPI; Address: 14482 Potsdam", HPI.toString()
        }
        // no redefinition in this layer for Employer class
        with (Layer.Employment) {
        	assertEquals "Name: HPI", HPI.toString()
        }

        Person somePerson = new Person(name: "Arvid Heise", address: "12345 Berlin", employer: HPI)
        // address and employment layer add additional information
        assertEquals "Name: Arvid Heise", somePerson.toString()
        with (Layer.Address) {
            assertEquals "Name: Arvid Heise; Address: 12345 Berlin", somePerson.toString()
        }
        with (Layer.Employment) {
            assertEquals "Name: Arvid Heise; [Employer] Name: HPI", somePerson.toString()
        }
        
        // depending of activation order this results in different order of added information
        with (Layer.Address) {
            with (Layer.Employment) {
            	// call stack would be:
        		// 1. [Layer.Employment] Person.toString(), invokes proceed()
        		// 2. [Layer.Address] Person.toString(), invokes proceed()
                // 3. Person.toString() returns "Name: Arvid Heise"
                // 2. [Layer.Address] Person.toString(), appends address and returns
                // 1. [Layer.Employment] Person.toString(), appends employer and returns
            	assertEquals "Name: Arvid Heise; Address: 12345 Berlin; [Employer] Name: HPI", somePerson.toString()
            }
        }
        
        with (Layer.Employment) {
            with (Layer.Address) {
            	// layer methods are invoked in reverse order
            	assertEquals "Name: Arvid Heise; [Employer] Name: HPI; Address: 12345 Berlin", somePerson.toString()
            }
        }
        // shorter form for previous nesting
        with ([Layer.Employment, Layer.Address]) {
            assertEquals "Name: Arvid Heise; [Employer] Name: HPI; Address: 12345 Berlin", somePerson.toString()
        }
        with (Layer.Employment + Layer.Address) {
            assertEquals "Name: Arvid Heise; [Employer] Name: HPI; Address: 12345 Berlin", somePerson.toString()
        }
        with (Layer.Employment, Layer.Address) {
            assertEquals "Name: Arvid Heise; [Employer] Name: HPI; Address: 12345 Berlin", somePerson.toString()
        }

        with (Layer.Address) {
        	// double activation changes activation order
            with (Layer.Employment + Layer.Address) {
            	assertEquals "Name: Arvid Heise; [Employer] Name: HPI; Address: 12345 Berlin", somePerson.toString()
                
            	// deactivation ensures that a specific layer is not activated
            	// independently of the number of previous activations
                without (Layer.Address) {
                    assertEquals "Name: Arvid Heise; [Employer] Name: HPI", somePerson.toString()
                }
            }
        }
    }
}